###############################################################
# R script for illustrating goodness of fit options in RSiena #
###############################################################

library(RSiena)        # add RSiena vocabulary to namespace
source("sienaGOF.R")   # add the sienaGOF function from package RSienaTest

# Load some of the MBA data:
advice1 <- as.matrix(read.table("MBA-Advice1.txt"))
advice2 <- as.matrix(read.table("MBA-Advice2.txt"))
perf <- as.matrix(read.table("MBA-Performance.txt"))[,1]

# Take a look at what's there now:
objects()
# What you see are data sets and goodness-of-fit related functions.

# Transform into RSiena session data:
advice <- sienaNet(array(c(advice1,advice2),dim=c(75,75,2)))
performance <- coCovar(perf)

# Bind together for an analysis in a sienaData object:
gof.data <- sienaDataCreate(advice,performance)

# Create sienaModel object containing estimation settings:
gof.model <- sienaModelCreate(cond=FALSE,projname='gof-estimates')

# Create an effects object containing the model specification:
gof.effects <- getEffects(gof.data)

# Optional: get first descriptives to double-check data:
# print01Report(gof.data,gof.effects,modelname='gof-descriptives')

# Estimate first an 'almost empty' model (just outdegree & reciprocity):
gof.results1 <- siena07(gof.model,data=gof.data,effects=gof.effects,
	returnDeps=TRUE)
# The last option guarantees that simulated data sets will be available
# in the resulting sienaFit object.

# Assess goodness of fit of the model based on these simulated networks:
gof.indegree1 <- sienaGOF(gof.results1,IndegreeDistribution,verbose=TRUE)
plot(gof.indegree1,key=0:8)
gof.outdegree1 <- sienaGOF(gof.results1,OutdegreeDistribution,verbose=TRUE)
plot(gof.outdegree1,key=0:8)
gof.geodesic1 <- sienaGOF(gof.results1,GeodesicDistribution,verbose=TRUE)
plot(gof.geodesic1,key=1:8)
gof.triads1 <- sienaGOF(gof.results1,TriadCensus,verbose=TRUE)
triad.keys <- c("003","012","102","021D","021U","021C","111D","111U",
		"030T","030C","201","120D","120U","120C","210","300")
plot(gof.triads1,key=triad.keys,center=TRUE,scale=TRUE)
# Results: Triad census has for 'more filled' triads completely bad fit.
#          Simulated geodesics are way too short.
#          Observed indegrees remain just outside 95% band of simulated ones.
#          Outdegree distribution almost fits well.

# Include triadic closure in the model to improve fit on both counts:
gof.effects <- includeEffects(gof.effects,transTrip)

# Estimate second model:
gof.results2 <- siena07(gof.model,data=gof.data,effects=gof.effects,
	returnDeps=TRUE)

# Again assess goodness of fit:
gof.indegree2 <- sienaGOF(gof.results2,IndegreeDistribution,verbose=TRUE)
plot(gof.indegree2,key=0:8)
gof.outdegree2 <- sienaGOF(gof.results2,OutdegreeDistribution,verbose=TRUE)
plot(gof.outdegree2,key=0:8)
gof.geodesic2 <- sienaGOF(gof.results2,GeodesicDistribution,verbose=TRUE)
plot(gof.geodesic2,key=1:8)
gof.triads2 <- sienaGOF(gof.results2,TriadCensus,verbose=TRUE)
plot(gof.triads2,key=triad.keys,center=TRUE,scale=TRUE)
# Results: Triad census acceptable fit except 021U, 021C, and 030C.
#          Simulated geodesics longer now but still too short.
#          Observed indegrees okay now except for isolates (indegree zero).
#          Outdegree distribution fits well.

# Include cyclical closure as well (two of the 'bad triads' contain a 'C'):
gof.effects <- includeEffects(gof.effects,cycle3)

# Estimate second model:
gof.results3 <- siena07(gof.model,data=gof.data,effects=gof.effects,
	returnDeps=TRUE)

# Again assess goodness of fit:
gof.indegree3 <- sienaGOF(gof.results3,IndegreeDistribution,verbose=TRUE)
plot(gof.indegree3,key=0:8)
gof.outdegree3 <- sienaGOF(gof.results3,OutdegreeDistribution,verbose=TRUE)
plot(gof.outdegree3,key=0:8)
gof.geodesic3 <- sienaGOF(gof.results3,GeodesicDistribution,verbose=TRUE)
plot(gof.geodesic3,key=1:8)
gof.triads3 <- sienaGOF(gof.results3,TriadCensus,verbose=TRUE)
plot(gof.triads3,key=triad.keys,center=TRUE,scale=TRUE)
# Results: Triad census still doubtful for 030C.
#          Fit on geodesic distribution remains unchanged, bad.
#          Isolates still a problem with the indegree distribution.
#          Outdegree distribution fits well.

# One more, final extension: add hierarchy determinant 'performance':
gof.effects <- includeEffects(gof.effects,egoX,altX,simX,interaction1="performance")

# Estimate second model:
gof.results4 <- siena07(gof.model,data=gof.data,effects=gof.effects,
	returnDeps=TRUE)

# Again assess goodness of fit:
gof.indegree4 <- sienaGOF(gof.results4,IndegreeDistribution,verbose=TRUE)
plot(gof.indegree4,key=0:8)
gof.outdegree4 <- sienaGOF(gof.results4,OutdegreeDistribution,verbose=TRUE)
plot(gof.outdegree4,key=0:8)
gof.geodesic4 <- sienaGOF(gof.results4,GeodesicDistribution,verbose=TRUE)
plot(gof.geodesic4,key=1:8)
gof.triads4 <- sienaGOF(gof.results4,TriadCensus,verbose=TRUE)
plot(gof.triads4,key=triad.keys,center=TRUE,scale=TRUE)
# Results: Except for the tail of the geodesic distribution, fit
#          is acceptable on the investigated criteria.

# Addendum: you can also investigate 'custom-made' functions as fit criteria:
source("CustomFit.R")

gof.descriptives4 <- sienaGOF(gof.results4,descriptives,verbose=TRUE)
plot(gof.descriptives4,key=descriptives.keys,center=TRUE,scale=TRUE)
# Result: On these macro features of the network, the model
#         performs not really well. Worst modelled is Krackhardt's
#         hierarchy measure and the size of the largest strong
#         component. Borderline are least upper boundedness,
#         the number of strong components, and media geodesic distance.